##############################################################################
# The top half is a kind of assembler to create Wang BASIC-2 $GIO programs.
# Each function provides easier to read mnemonics than raw hex.
#
# Skip down to "main program" to see how it is used. Running this python
# program generates a BASIC-2 program which, when run, prints all the
# prime numbers < 65536 (though it gets slower and slower as it goes).
# The trick here is that $GIO does not support even add or subtract, much
# less multiplication/division/square root, yet it accomplishes the task.
#
#    1. py -3 primes_gio.py > primes_gio.w22
#    2. from the emulator, File/Script... and pick primes_gio.w22
#    3. RUN
#
# Author: Jim Battle
# Date:   2020/03/07

################################# assembler #################################

# offset of current instruction
curr_pc = 0

# the generated program instructions
program = []

# key is a label string, and the value is the offset of the pc at the time
# of the label definition
label_vals = {}

# list of entries which have a branch to a label
# these are used to patch branches at the very end, as there might have
# been forward references. they keys are pc offsets, and the value is the
# label name.
label_refs = {}

# creates a label at the current PC
def make_label(label_str):
    if label_str in label_vals.keys():
        assert False, "Label %s defined twice" % label_str
    label_vals[label_str] = curr_pc

# save instruction to curr_pc, increment curr_pc
def emit(instr):
    global curr_pc
    assert 0x0000 <= instr <= 0xffff
    program.append(instr)
    curr_pc += 1

# valid register values are 0-15
def valid_regnum(regnum):
    return 0 <= regnum <= 15;

def valid_imm(immval):
    return 0x00 <= immval <= 255;

# set register r to immediate 8b value: 0rh1h2
def movi(r, immval):
    assert valid_regnum(r)
    assert valid_imm(immval)
    instr = 0x0000 | (r << 8) | immval
    emit(instr)

# set register pair r,r+1 to immediate 16b value
def movix(r, immval):
    movi(r+0, immval // 0xff)
    movi(r+1, immval &  0xff)

# move contents of register r1 to register r2: 11r1r2
def mov(r2, r1):
    assert valid_regnum(r1)
    assert valid_regnum(r2)
    instr = 0x1100 | (r1 << 4) | r2
    emit(instr)

# move contents of register pair r1,r1=1 to register r2,r2+1
def movx(r2, r1):
    mov(r2+0,r1+0)
    mov(r2+1,r1+1)

# increment or decrement 8b register
def inc(r, delta=1):
    assert valid_regnum(r)
    instr = 0x1900 | (r << 4)
    if delta == 1: instr |= 0
    elif delta == 2: instr |= 1
    elif delta == -1: instr |= 3
    elif delta == -2: instr |= 2
    else: assert False, "bad inc(r, %d)" % delta
    emit(instr)

def dec(r, delta=1): inc(r, -delta)

# increment or decrement 16b register pair
def incx(r, delta=1):
    assert valid_regnum(r)
    instr = 0x1900 | (r << 4)
    if delta == 1: instr |= 4
    elif delta == 2: instr |= 5
    elif delta == -1: instr |= 7
    elif delta == -2: instr |= 6
    else: assert False, "bad incx(r, %d)" % delta
    emit(instr)

def decx(r, delta=1): incx(r, -delta)

# no-operation
def noop():
    mov(0, 0)  # writes to r0 are ignored

# set condition code true
def cctrue():
    instr = 0x1000
    emit(instr)

# set condition code if r1 == r2
def eq(r1, r2):
    assert valid_regnum(r1)
    assert valid_regnum(r2)
    instr = 0x1C00 | (r1 << 4) | r2
    emit(instr)

# set condition code if register pair r1,r1+1 == r2,r2+1
def eqx(r1, r2):
    assert valid_regnum(r1)
    assert valid_regnum(r2)
    instr = 0x1D00 | (r1 << 4) | r2
    emit(instr)

# set condition code if r1 > r2
def gt(r1, r2):
    assert valid_regnum(r1)
    assert valid_regnum(r2)
    instr = 0x1E00 | (r1 << 4) | r2
    emit(instr)

# set condition code if register pair r1,r1+1 > r2,r2+1
def gtx(r1, r2):
    assert valid_regnum(r1)
    assert valid_regnum(r2)
    instr = 0x1F00 | (r1 << 4) | r2
    emit(instr)

# branch to specified microcommand if condition code true
def jt(label_str):
    instr = 0xD000
    label_refs[curr_pc] = label_str
    emit(instr)

# branch to specified microcommand if condition code is false
def jf(label_str):
    instr = 0xE000
    label_refs[curr_pc] = label_str
    emit(instr)

# send 8b value via OBS to currently addressed device
# either in 8b integer or a single character is acceptable
def echo(immval):
    if isinstance(immval,str):
        assert len(immval) == 1
        immval = ord(immval)
    assert valid_imm(immval)
    instr = 0x4000 | immval
    emit(instr)

# send register r via OBS to currently addressed device
def echor(r):
    assert valid_regnum(r)
    instr = 0x4200 | (r << 4)
    emit(instr)

# point to n-th data buffer; reset data count
def buffer(n):
    assert 0x00 <= n <= 0xff
    instr = 0x1800 | n
    emit(instr)

# write data buffer to register r; increment data pointer
# this read/write terminology is confusing, but is what wang docs use
def wrbuf(r):
    assert valid_regnum(r)
    instr = 0x1B01 | (r << 4)
    emit(instr)

# read register r and save in data buffer; increment data pointer
# this read/write terminology is confusing, but is what wang docs use
def rdbuf(r):
    assert valid_regnum(r)
    instr = 0x1B02 | (r << 4)
    emit(instr)

# fix up program now that all label offsets are known
def resolve_labels():
    for pc in label_refs.keys():  # pc is instruction referencing a label
        label = label_refs[pc]    # this is the label it named
        if label in label_vals:
            program[pc] |= label_vals[label]
        else:
            assert False, "pc 0x%04x refers to label %s, but it isn't defined" % (pc, label)

# print out the generated program
def dump_program():
    # pad noops at the to make a mulitple of 8
    while (len(program) % 8) != 0:
        program.append(0x0000)
    steps = len(program)
    print('CLEAR')
    print('10 DIM P$(%d)16' % (steps//8))
    for pc in range(0,len(program)):
        if (pc % 8 == 0):
            print('%d P$(%d)=HEX(' % ( 20+10*(pc/8), 1+(pc/8)), end='')
        print('%04X' % program[pc], end='')
        if (pc % 8 == 7):
            print(')')
        else:
            print(' ', end='')
    print('%d $GIO /005 (P$(),R$) D$' % (20 + 10*(steps//8)) )
    #print('%d PRINT "candidate=";D$' % (30 + 10*(steps//8)) )
    #print('%d PRINT "candidate=";:HEXPRINT STR(R$,2,2)' % (40 + 10*(steps//8)) )
    #print('%d PRINT "square   =";:HEXPRINT STR(R$,4,2)' % (50 + 10*(steps//8)) )
    #print('%d PRINT "sqinc    =";:HEXPRINT STR(R$,6,2)' % (60 + 10*(steps//8)) )
    #print('%d PRINT "div_limit=";:HEXPRINT STR(R$,11,1)' % (70 + 10*(steps//8)) )

############################ main program ############################
# the pseudocode is roughly this:
#
#    initialize:
#        candidate(16) = 1
#        square(16)    = 4
#        sqinc(16)     = 5
#        limit(8)      = 2
#        save_ascii "00001"
#
#    next_candidate:
#        candidate += 2
#        if (candidate > 0xffff)
#            goto done
#
#        read ascii
#        update ascii += 2
#        save ascii
#        echo ascii
#
#        if (candidate > square)
#           square += sqinc      4 -> 9 -> 16 -> 25 -> etc
#           sqinc += 2           5 -> 7 ->  9 -> 11 -> etc
#           limit += 1           2 -> 3 ->  4 ->  5 -> etc
#
#        divisor(8) = 1
#
#    next_divisor:
#        divisor += 2
#        if (divisor > limit)
#           goto is_prime
#        remainder(16) = candidate
#
#    subtract_divisor:
#        if (divisor > remainder)
#            goto next_divisor
#        loop_count(8) = divisor
#    dec_loop:
#        remainder -= 1
#        loop_count -= 1
#        if (loop_count > 0)
#            goto dec_loop
#        if (remainder == 0)  # evenly divisible
#            goto next_candidate
#        goto subtract_divisor
#
#    is_prime:
#        print cr/lf
#        goto next_candidate
#
#    done:
#        noop

# register use:
zero         = 0   # r0 is always 0x00 on read, ignored on write

divisor      = 0   # r0,r1 -- note r0 is always 0x00
divisor_ms   = 0
divisor_ls   = 1

candidate    = 2   # r2,r3
candidate_ms = 2
candidate_ls = 3

square       = 4   # r4,r5: div_limit**2

sqinc        = 6   # r6,r7: the amount to increment square to get the next one
sqinc_ms     = 6
sqinc_ls     = 7

# r9,r10: data buffer offset, affected by buffer(), rdbuf(), and wrbuf()

div_limit    = 11  # floor(sqr(candidtate)) -- the largest divisor to test

remainder    = 12   # r12,r13: candidate value remainder while dividing
remainder_ms = 12
remainder_ls = 13

loop_count   = 14   # performs subtraction through repetitive decrement loop

tmp          = 15   # holds temporary byte values for comparison

# this temporarily hold the ascii representation of the candidate value
# we reuse registers which are unused in this phrase of the algorithm
asc_1        = 1
asc_10       = 8    # care is needed here, as r8 has special meaning
asc_100      = 12
asc_1k       = 13
asc_10k      = 14

#### print intro message
for ch in 'Prime numbers:': echo(ch)
echo(0x0d); echo(0x0a)

echo('2')
echo(0x0d); echo(0x0a)

#### initialize registers

movix(candidate, 1)   # value to test
movi (div_limit, 2)   # at first don't test divisor greater than 2
movix(square,    4)   # div_limit**2
movix(sqinc,     5)   # amount to increment square to produce the next one

# fill data buffer with '00001'
buffer(0)
movi(tmp,0x30); rdbuf(tmp)
movi(tmp,0x30); rdbuf(tmp)
movi(tmp,0x30); rdbuf(tmp)
movi(tmp,0x30); rdbuf(tmp)
movi(tmp,0x31); rdbuf(tmp)

#### generate all odd numbers: 3, 5, 7, 9, 11, 13, ...
make_label("next_candidate")
incx(candidate)
eq(candidate_ms, zero); jf('nc2')
eq(candidate_ls, zero); jt('done')   # we wrapped past 16b!
make_label('nc2')
incx(candidate)

#### incrementally calculate divisor limit needed for this candidate value
gtx(candidate, square); jf('inc_ascii')
movx(remainder, sqinc)   # use remainder as a 16b count value
incx(sqinc, 2)           # 5 -> 7 -> 9 -> 11 -> etc
inc(div_limit)
make_label('bump_square')
incx(square)
decx(remainder)
eq(remainder_ls, zero); jf('bump_square')
eq(remainder_ms, zero); jf('bump_square')

##### read all five ascii bytes into registers
make_label('inc_ascii')
buffer(0)  # reset the pointer
wrbuf(asc_10k)  # ms
wrbuf(asc_1k)
wrbuf(asc_100)
wrbuf(asc_10)
wrbuf(asc_1)  # ls

# increment units place by two
movi(tmp,0x3B)  # '9'+2
inc(asc_1, 2)
eq(asc_1, tmp)
jf('save_count')
movi(asc_1, 0x31)  # '1'

# increment 10s
movi(tmp, 0x3A)  # '9'+1
inc(asc_10)
eq(asc_10, tmp)
jf('save_count')
movi(asc_10, 0x30)  # '0'

# increment 100s
inc(asc_100)
eq(asc_100, tmp)
jf('save_count')
movi(asc_100, 0x30)  # '0'

# increment 1000s
inc(asc_1k)
eq(asc_1k, tmp)
jf('save_count')
movi(asc_1k, 0x30)  # '0'

# increment 10000s
inc(asc_10k)   # we don't worry about wrapping, since 65535 is max count

# save back to buffer
make_label('save_count')
buffer(0)
rdbuf(asc_10k)
rdbuf(asc_1k)
rdbuf(asc_100)
rdbuf(asc_10)
rdbuf(asc_1)

# echo current test candidate to crt
movi(tmp, 0x0D)
echor(tmp)
movi(tmp, 0x30)  # '0'
eq(asc_10k, tmp); jf('emit_cnt_10000')
eq(asc_1k,  tmp); jf('emit_cnt_1000')
eq(asc_100, tmp); jf('emit_cnt_100')
eq(asc_10,  tmp); jf('emit_cnt_10')
jf('emit_cnt_1')
make_label('emit_cnt_10000')
echor(asc_10k)
make_label('emit_cnt_1000')
echor(asc_1k)
make_label('emit_cnt_100')
echor(asc_100)
make_label('emit_cnt_10')
echor(asc_10)
make_label('emit_cnt_1')
echor(asc_1)

# test just printing all odd numbers up to 65535:
#echo(0x0A)
#jf('next_candidate')

##### do a test division of all odd numbers below floor(sqr(candidate))
# - it could be optimized by saving the generated primes into an
#   output data array, and then scanning that array for test divisors,
#   instead of also testing compound values. for the range of numbers
#   being tested here, the speedup would be 2x to 2.5x.
#   there are 53 odd primes < 256.
#
# - the real speed killer is that for large candidate numbers, testing
#   for divisibility is very very slow. for example, testing if 8191 is
#   prime requires decrementing the candidate number 8191 times inside
#   a nested loop doing 3 decrements at a time. it turns out it is a prime,
#   so it isn't. the algorithm then decrements the candidate number five
#   at a time, again 8191 times total to discover the number isn't divisible
#   by 5, and so on.
#
#   a bigger improvement would be to also generate k*prime and save it too.
#   what is the advantage? If the candidate number is > 255, and (k*prime)
#   is also bigger than 255, rather than decrementing by prime over and over,
#   instead the remainder can be decremented by (k*prime mod 256), then the
#   ms 8 bits can be decremented by floor(k*prime / 256). after that, then
#   the old fashioned reduction by 'prime' decrements at a time can be used.
#
#   probably the best approach would be to update the k*prime values each
#   time the list is scanned for a new candidate. essentially read out k*prime,
#   in preparation for the divisibility test, compute (k*prime + prime); if
#   that is > candidate, it isn't divisible; if it is == candidate, then it
#   is divisible, otherwise save that updated (k+1)*prime back to the list.

# NB: the first version of this program didn't apply square root limit,
# and instead tested all odd numbers < candidate and also < 256.  I timed
# that version vs this version for computing all primes < 1000.
#    simple: 24m 15s
#      this:  4m 20s
#   basic-2:     <8s   (see primes_basic2.w22)
# the speed advantage would reduce as the numbers approach 64K, since the
# square root limit would begin to approach 255 anyway.

movi(divisor_ls, 1)
make_label("next_divisor")
inc(divisor_ls, 2)  # next odd

gt(divisor_ls, div_limit); jt('is_prime')

# see if divisor can be subtracted evenly from test candidate
movx(remainder, candidate)

# subtract divisor N by decrementing N times
make_label('subtract_divisor')
gtx(divisor, remainder); jt('next_divisor')  # divisor is greater than remainder
mov(loop_count, divisor_ls)
make_label('dec_loop')
decx(remainder)
dec(loop_count)
eq(loop_count, zero); jf('dec_loop')

# remainder has been reduced by the test divisor.
# if it is non-zero, try subtracting the divisor again
make_label('subtract_divisor_done')
eq(remainder_ms, zero); jf('subtract_divisor')
eq(remainder_ls, zero); jf('subtract_divisor')
# it was evenly divisible
jf('next_candidate')

# we've found a prime number; emit linefeed
make_label('is_prime')
echo(0x0a)
jf('next_candidate')

# exhausted our 16b counters, so quit
make_label('done')
noop()

resolve_labels()
dump_program()
